"""Module grepenv provides functionality for greping through the environment.
"""
import re
import os

from typing import List

from dataclasses import dataclass


@dataclass
class EnvItem:
    """A single key/value pair entry in the environment."""

    key: str
    value: str


def parse_environment() -> List[EnvItem]:
    """Extract all key/value pairs from the environment."""
    env_items = [EnvItem(k, v) for k, v in os.environ.items()]
    return sorted(env_items, key=lambda e: e.key)


def filter_env_by_regular_expression(
    pat: re.Pattern, keys_only: bool = False, values_only: bool = False
) -> List[EnvItem]:
    """Filters the environment for key/value pairs that match the given regular
    expression.
    """

    filter_fn = lambda e: (pat.search(e.key) and not values_only) or (
        pat.search(e.value) and not keys_only
    )

    return [e for e in parse_environment() if filter_fn(e)]



import typer

from grepenv.cli import (
    print_environment,
    print_matching_keys,
    show_examples,
    try_compile_regex_pattern,
)

from .grepenv import filter_env_by_regular_expression, parse_environment


app = typer.Typer(add_completion=False, no_args_is_help=True)

_HELP_STRING = """
greps the env

\b
By default, all keys and values are searched for matches.
See options to specify only keys, or only values.

\b
Pattern matching is done using regex SEARCH. Use anchor characters if 
matching the whole key or value is necessary.
"""

_EPILOG_STRING = """
Call grepenv with the --example flag to see some example usage.
"""


@app.command(help=_HELP_STRING, epilog=_EPILOG_STRING, no_args_is_help=True)
def _(
    pattern: str = typer.Argument(
        ".*", help="Regular expression pattern to search with."
    ),
    respect_case: bool = typer.Option(
        False,
        "-c",
        "--respect-case",
        help="Respect case of pattern characters. By default, the given regular expression will be set to ignore the case of alphabetic characters.",
    ),
    keys_only: bool = typer.Option(False, "-k", "--keys", help="Only search keys."),
    values_only: bool = typer.Option(
        False, "-v", "--values", help="Only search values."
    ),
    find_key: bool = typer.Option(
        False,
        "-fk",
        "--find-key",
        help="Modified behavior of grepenv. Will grep for all keys that match the given pattern, and return their corresponding value with no formatting, one per line.",
    ),
    no_highlight: bool = typer.Option(
        False, "-nh", "--no-highlight", help="Disable match highlighting."
    ),
    example: bool = typer.Option(False, "--example", help="Print some example usage."),
):
    # Exit on example
    if example:
        return show_examples()

    pat = try_compile_regex_pattern(pattern, ignore_case=not respect_case)

    # Handle find key branch
    if find_key:
        return print_matching_keys(parse_environment(), pat)

    # Regular highlight and print functionality
    env = filter_env_by_regular_expression(
        pat, keys_only=keys_only, values_only=values_only
    )

    print_environment(
        env,
        pat,
        keys_only=keys_only,
        values_only=values_only,
        highlight=not no_highlight,
    )


def main():
    app()


if __name__ == "__main__":
    main()



"""Helper functionality for the cli.

These functions may call sys.exit(), and so should only be called form the
cli itself.
"""
import re
import sys

from typing import NoReturn, Union, List

from rich.console import Console
from rich.live import Live
from rich.syntax import Syntax
from rich.panel import Panel

from grepenv.grepenv import EnvItem


_CONSOLE = Console()


def try_compile_regex_pattern(
    pattern: str, ignore_case: bool = True
) -> Union[re.Pattern, NoReturn]:
    """Attempt to compile the given regex `pattern`.

    If an exception is thrown during compilation, an error will be printed,
    and the program will exit.
    """
    try:
        return re.compile(pattern, re.IGNORECASE if ignore_case else 0)
    except Exception as e:
        print_error(f"Couldn't compile regular expression ({e})")
        sys.exit(1)


def highlight_string(var: str, pat: re.Pattern) -> str:
    """Scan the given string variable, and return a modified version in which
    all ranges matched by `pat` surrounded in `rich` highlighting tags.
    """
    if not pat.search(var):
        return var

    for m in reversed(list(pat.finditer(var))):
        start = m.start()
        end = m.end()

        s0 = var[:start]
        s1 = var[start:end]
        s2 = var[end:]

        var = f"{s0}[red]{s1}[/]{s2}"

    return var


def print_environment(
    env: List[EnvItem],
    pat: re.Pattern,
    keys_only: bool = False,
    values_only: bool = False,
    highlight: bool = True,
):
    """Prints all environment items with styling according to the given options."""
    for x in env:
        # Format key
        if values_only:
            key_s = f"[dim]{x.key}[/]"
        elif highlight:
            key_s = highlight_string(x.key, pat)
        else:
            key_s = x.key

        # Format value
        if keys_only:
            value_s = f"[dim]{x.value}[/]"
        elif highlight:
            value_s = highlight_string(x.value, pat)
        else:
            value_s = x.value

        # concat
        _CONSOLE.print(f"{key_s}={value_s}", highlight=False)


def print_matching_keys(env: List[EnvItem], pat: re.Pattern):
    for x in env:
        if pat.search(x.key):
            print(x.value)


def print_error(m: str):
    """Print a formatted error message `m`."""
    _CONSOLE.print(f"[bold error]ERROR: [/]{m}")


def show_examples():
    examples = [
        (
            "Find all environment keys and variables containing the letters 'xdg'",
            """
            $ grepenv xdg
            XDG_CONFIG_DIRS=/etc/xdg/xdg-cinnamon:/etc/xdg
            XDG_CURRENT_DESKTOP=X-Cinnamo
            XDG_RUNTIME_DIR=/run/user/1000
            XDG_SEAT=seat0
            ...
            """,
        ),
        (
            "Find all environment keys and variables containing the letters 'xdg' (strict lowercase)",
            """
            $ grepenv xdg -c
            XDG_CONFIG_DIRS=/etc/xdg/xdg-cinnamon:/etc/xdg
            """,
        ),
        (
            "Find all keys containing the letters 'git'",
            """
            $ grepenv git -k
            GITHUB_API_TOKEN=abc_NlNhalNDL78NAhdKhNAk78bdf7f
            """,
        ),
        (
            "Extract all values from keys which contain the letters 'git'",
            """
            $ grepenv git -fk
            abc_NlNhalNDL78NAhdKhNAk78bdf7f
            """,
        ),
        (
            "Get all environment that looks like an api key",
            """
            $ grepenv "_api_(key|token)_" -k
            GITHUB_API_TOKEN=abc_NlNhalNDL78NAhdKhNAk78bdf7f
            OPENAI_API_KEY=123_abcdefghijklmnop
            """,
        ),
    ]

    for title, body in examples:
        _CONSOLE.print(Panel(Syntax(body, "bash"), title=title, title_align="left"))
